#include "Common.hpp"
#include "Search.hpp"
#include "TrialDivision.hpp"
#include "MillerRabin.hpp"

namespace {

  const int TAG_NEXT = 0;
  const int TAG_FINISH = 1;
  
  unsigned MpiMaster(mpz_t rop, unsigned size, unsigned long seed) {
    unsigned processed = 0;
    int bufferSize = (size + 7) / 8;
    unsigned char *buffer = new unsigned char[(size + 7) / 8];
    MPI_Status status;
    gmp_randstate_t rand;
    mpz_t number;
    
    gmp_randinit_default(rand);
    gmp_randseed_ui(rand, seed);
    mpz_init(number);

    for (;;) {
      MPI_Recv(buffer, bufferSize, MPI_UNSIGNED_CHAR, MPI_ANY_SOURCE,
        MPI_ANY_TAG, MPI_COMM_WORLD, &status);
      if (status.MPI_TAG == TAG_NEXT) {
        RandomNumber(number, rand, size);
        ++processed;
        mpz_export(buffer, NULL, -1, 1, -1, 0, number);
        MPI_Send(buffer, bufferSize, MPI_UNSIGNED_CHAR, status.MPI_SOURCE,
          TAG_NEXT, MPI_COMM_WORLD);
      } else if (status.MPI_TAG == TAG_FINISH) {
        mpz_import(rop, bufferSize, -1, 1, -1, 0, buffer);
        int size;
        MPI_Comm_size(MPI_COMM_WORLD, &size);
        for (int i = 1; i < size; ++i) {
          if (i != status.MPI_SOURCE) {
            MPI_Send(NULL, 0, MPI_INT, i, TAG_FINISH, MPI_COMM_WORLD);
          }
        }
        break;
      }
    }
    gmp_randclear(rand);
    mpz_clear(number);
    return processed;
  }

  void MpiSlave(unsigned size, unsigned long seed) {
    int bufferSize = (size + 7) / 8;
    unsigned char *buffer = new unsigned char[(size + 7) / 8];
    MPI_Status status;
    gmp_randstate_t mrRand;
    mpz_t number;
    TrialDivision trial;
    MillerRabin mr;

    gmp_randinit_default(mrRand);
    gmp_randseed_ui(mrRand, seed);
    mpz_init(number);

    for (;;) {
      MPI_Send(NULL, 0, MPI_INT, 0, TAG_NEXT, MPI_COMM_WORLD);
      MPI_Recv(buffer, bufferSize, MPI_UNSIGNED_CHAR, 0, MPI_ANY_TAG,
        MPI_COMM_WORLD, &status);

      if (status.MPI_TAG == TAG_NEXT) {
        mpz_import(number, bufferSize, -1, 1, -1, 0, buffer);
        if (trial.Test(number) && mr.Test(number, mrRand)) {
          MPI_Send(buffer, bufferSize, MPI_UNSIGNED_CHAR, 0, TAG_FINISH,
            MPI_COMM_WORLD);
          break;
        }
      } else if (status.MPI_TAG == TAG_FINISH) {
        break;
      }
    }
    gmp_randclear(mrRand);
    mpz_clear(number);
  }

}

unsigned MpiSearch(mpz_t rop, unsigned size, unsigned long seed,
                   int *argc, char **argv[]) {
  int rank;
  MPI_Init(argc, argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  unsigned processed;
  if (rank == 0) {
    processed = MpiMaster(rop, size, seed);
  } else {
    MpiSlave(size, seed);
    processed = 0;
  }
  MPI_Finalize();
  return processed;
}
