#include "Common.hpp"
#include "Search.hpp"
#include "TrialDivision.hpp"
#include "MillerRabin.hpp"

namespace {
  struct Shared {
    unsigned size;
    unsigned long seed;
    unsigned processed;

    gmp_randstate_t genRand;
    boost::mutex genMutex;

    mpz_t result;
    boost::mutex resultMutex;
    boost::condition_variable resultCond;
  };

  void Worker(Shared *shared) {
    gmp_randstate_t mrRand; 
    mpz_t number;
    TrialDivision trial;
    MillerRabin mr;

    gmp_randinit_default(mrRand);
    gmp_randseed_ui(mrRand, shared->seed);
    mpz_init(number);

    while (!boost::this_thread::interruption_requested()) {
      {
        boost::lock_guard<boost::mutex> lock(shared->genMutex);
        RandomNumber(number, shared->genRand, shared->size);
        ++shared->processed;
      }
      if (trial.Test(number) && mr.Test(number, mrRand)) {
        boost::unique_lock<boost::mutex> lock(shared->resultMutex);
        mpz_swap(shared->result, number);
        shared->resultCond.notify_one();
        break;
      }
    };
    mpz_clear(number);
  }
}

unsigned ThreadedSearch(mpz_t rop, unsigned size, unsigned long seed,
                        unsigned nThreads) {
  if (nThreads == 0) {
    nThreads = boost::thread::hardware_concurrency();
    if (nThreads == 0) {
      nThreads = 1;
    }
  }

  Shared shared;
  shared.size = size;
  shared.seed = seed;
  shared.processed = 0;
  gmp_randinit_default(shared.genRand);
  gmp_randseed_ui(shared.genRand, seed);
  mpz_init(shared.result);

  boost::thread_group workers;
  for (unsigned i = 0; i < nThreads; ++i) {
    boost::thread *thread = new boost::thread(&Worker, &shared);
    workers.add_thread(thread);
  }
  boost::unique_lock<boost::mutex> lock(shared.resultMutex);
  shared.resultCond.wait(lock);
  workers.interrupt_all();
  workers.join_all();
  shared.resultMutex.unlock();
  
  mpz_swap(rop, shared.result);
  gmp_randclear(shared.genRand);
  mpz_clear(shared.result);
  return shared.processed;
}
